/*
   This file is part of the Aulib library.
   Name: afplayer.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AF_PLAYER_H_
#define _AF_PLAYER_H_

#include "audata.h"
#include <string>
using namespace std;

#ifdef AF
// ====================== using AF ==========================
#include <AF/AFlib.h>
#include <AF/AFUtils.h>

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      AFPlayer

/**
 * AFPlayer represents a tool which plays audio files
 * (class AudioData) on an AF-server.
 *
 * @see AudioData
 * @author Silvia Pfeiffer, Peter Tomczyk
 * @version $Id: afplayer.h,v 1.8 1998/09/28 10:48:05 tomczyk Exp $
 */
class AFPlayer {

  public:

    /**
     * creates new AFPlayer connetced to the specified
     * server and playing on the given device.
     *
     * @param device audio-device number (id)
     * @param server server name
     * @exception Exception
     */
    AFPlayer(int device = 0, const string& server = "");

    /// destroys AFPlayer.
    virtual ~AFPlayer();

    /// plays audata from the current position to the end.
    void play(const AudioData& audata);

    /// plays 'length' samples form the current position.
    void play(const AudioData& audata, long int length);

    /// sets the loudness.
    void gain(int gain) { _gain = gain; }

    /**
     * sets the time space between the call of play() and
     * starting playing the file.
     *
     * @param delay time in seconds
     */
    void delay(float delay) { _delay = delay; }

    /// returns the loudness.
    int gain() const { return _gain; }

    /// return the delay.
    float delay() const { return _delay; }

    /// returns the id of the first device on the server.
    int firstDevice() const { return 0; }

    /// returts the id of the last device on the server + 1.
    int lastDevice() const { return _aud->ndevices; }

  private:

    /// calculates the size of an AF unit.
    int calcUnitSize(const AudioData& audata) const;

    /// calculates the buffer size in seconds.
    int calcPlayBufSize(const AudioData& audata, AC ac) const;

    /// calculates how many times server must be called to play samples.
    long int calcLoops(const AudioData& audata, long int length) const;

    /// sets starting time from current AF audio device time
    void setStartingTime(AC ac);

    /// plays audata using the given audio context.
    void playFile(const AudioData& audata, AC ac, long int loops);

    /// creates audio context for the given AudioData.
    AC createACFor(const AudioData& audata);

    /// finds a suitable device.
    ADevice findDeviceFor(const AudioData& audata) const;

    /// returns a set of attributes needed to create audio context.
    AFSetACAttributes getAttributesFor(const AudioData& audata) const;

    /// creates audio context using the given arguments.
    AC createAC(const AudioData& audata, ADevice device, AFSetACAttributes attrib) const;

    /**
     * finds suitable default device for audata if the choosen 
     * device cannot be used. Returns the first device not
     * connected to the phone.
     */
    ADevice findDefaultDeviceFor(const AudioData& audata) const;

    /// destroys audio context.
    void deleteAC(AC ac);

    AFPlayer(const AFPlayer&);
    AFPlayer& operator=(const AFPlayer&);

    /// audio-server name.
    string _server;

    /// audio-device (id).
    int _device;

    /// loudness.
    int _gain;

    /// time space between the call of play() and starting playing.
    float _delay;

    /// used audio context.
    AFAudioConn* _aud;

    /// AF timing.
    ATime _time;

    /// buffer size, in hundreth seconds.
    static float BUFSIZE;
};

////////////////////////////////////////////////////////////

inline
AFPlayer::~AFPlayer() { AFCloseAudioConn(_aud); }

////////////////////////////////////////////////////////////

inline
int AFPlayer::calcUnitSize(const AudioData& audata) const {
  return BytesPerUnit(audata.AFEncodeType());
}

////////////////////////////////////////////////////////////

inline
int AFPlayer::calcPlayBufSize(const AudioData& audata, AC ac) const {
  return static_cast<int>(ac->device->playSampleFreq /
    SampsPerUnit(audata.AFEncodeType()) * BUFSIZE);
}

////////////////////////////////////////////////////////////

inline
void AFPlayer::deleteAC(AC ac) {
  delete ac;
}

// ====================== using AF ==========================
#else
// ===================== without AF =========================

class AFPlayer {
  public:
    AFPlayer(int device = 0, const string& server = "");

    void play(const AudioData&);
    void play(const AudioData&, long int length);

    void gain(int gain) { _gain = gain; }
    void delay(float delay) { _delay = delay; }

    int gain() const { return _gain; }
    float delay() const { return _delay; }
    int firstDevice() const { return 0; }
    int lastDevice() const { return 0; }

  private:
    string _server;
    int _device;
    int _gain;
    float _delay;
};

AFPlayer::AFPlayer(int device = 0, const string& server = "")
  : _server(server), _device(device), _gain(0), _delay(0.1) { }

void AFPlayer::play(const AudioData&) {
  throw Exception("AFPlayer::play() - Playing files is not supported.");
}

void AFPlayer::play(const AudioData&, long int length) {
  throw Exception("AFPlayer::play() - Playing files is not supported.");
}

// ===================== without AF =========================
#endif

#endif
