/*
   This file is part of the Aulib library.
   Name: aucut.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AU_CUT_
#define _AU_CUT_

#include "autypes.h"
#include "error.h"
#include "fft.h"
#include "aucache.h"
#include <cstddef>
#include <cmath>
#include <iostream.h>
using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                abstract CutDetector

/**
 * CutDetector is the abstract base class of all tools, which
 * perform cut detection on audio files. CutDetector works
 * iterative. It takes windows of audio data one by one
 * (Method cutOccurres(AudioCache)) and it returns true if 
 * a cut occurres between the current window and the previous
 * one. Otherwise it returns false. Example:
 * <pre>
 *   // 'audata' is of AudioData
 *   // 'detector' is a CutDetector
 *   bool cut;
 *   while(audata.load(win_size) == win_size) {
 *     cut = detector.cutOccurres(audata.cache());
 *     cout << "Cut? " << cut << endl;
 *   }
 * </pre>
 * <p>
 * CutDetector uses an indicator to calculate cuts. It may be
 * amplitude, window energy, frequency spectrum or something 
 * else. CutDetector implements all parts of the detection 
 * algorithm except how to calculate the indicator and how to 
 * compare two indicators. A concrete detector class must be a 
 * subclass of CutDetector and it must override the pure
 * virtual functions analyse() and compare().
 * <p>
 * The data type of the indicator is implemented as a
 * template parameter of CutDetector. So it can be defined
 * in the subclass declaration. Example:
 * <pre>
 *  class CutDetectorUsingEnergy : public CutDetector<double> {
 *     ...
 *    void analyse(const AudioCache& aucache) {
 *      *_curIndicator = energy(aucache); -> returns double
 *    }
 *     ...
 *  };
 * </pre>
 * <p>
 * Indicator must define the following functions:
 * <ul>
 * <li>Default constructor
 * <li>Copy constructor
 * <li>Destructor
 * <li>Assignment operator=(Indicator)
 * <li>operator*(Float, Indicator)
 * <li>operator+(Indicator, Indicator)
 * </ul>
 * operator* and operator+ are only used by adjust().
 * If they are not defined adjust() must be 
 * overriden.
 *
 * @memo CutDetector is the abstract base class of all tools,
 *       which perform cut detection on audio files
 *
 * @author Christoph Gerum, Silvia Pfeiffer, Peter Tomczyk
 * @version $Id: aucut.h,v 1.13 1998/09/11 11:09:57 tomczyk Exp $
 * @see IndicatorCompare
 */
template<class Indicator>
class CutDetector {
  friend ostream& operator<<(ostream& os, const CutDetector& det) {
    return det.printOn(os);
  }
  public:
    CutDetector();
    CutDetector(unsigned int nparts, Float weight);
    virtual ~CutDetector() = 0;

    virtual bool cutOccurs(const AudioCache& aucache);
    void reset() { _historyExists = false; _simCount = 0; }
    void simParts(unsigned int num) { _simParts = num; }
    void weight(Float weight) { _weight = weight; }

    unsigned int simParts() const { return _simParts; }
    Float weight() const { return _weight; }

  protected:
    enum Similarity { EQUAL, SIMILAR, UNEQUAL };
    virtual Indicator* analyse(const AudioCache& aucache) = 0;
    virtual Similarity compareTo(const Indicator& curInd) = 0;
    virtual bool decide(Similarity sim);
    virtual void adjust(const Indicator& curInd, bool cut);
    void initHistory(const Indicator& ind);

    virtual ostream& printOn(ostream& os) const;

    unsigned int _simParts;
    unsigned int _simCount;
    Float _weight;
    bool _historyExists;
    Indicator _history;

  private:
    CutDetector(const CutDetector&);
    CutDetector& operator=(const CutDetector&);
};

////////////////////////////////////////////////////////////

template<class Indicator>
CutDetector<Indicator>::CutDetector(unsigned int nparts, Float weight)
  : _simParts(nparts), _simCount(0), _weight(weight), 
    _historyExists(false) { }

////////////////////////////////////////////////////////////

template<class Indicator>
CutDetector<Indicator>::CutDetector() : _simParts(1), 
  _simCount(0), _weight(0.5), _historyExists(false) { }

////////////////////////////////////////////////////////////

template<class Indicator>
inline CutDetector<Indicator>::~CutDetector() { }

////////////////////////////////////////////////////////////

template<class Indicator>
bool CutDetector<Indicator>::cutOccurs(const AudioCache& aucache) {

  Similarity sim;
  bool cut;
  Indicator* ind;

  ind = analyse(aucache);

  if (!_historyExists) {
    initHistory(*ind);
    return true;
  }
  else {
    sim = compareTo(*ind);
    cut = decide(sim);
    adjust(*ind, cut);
    return cut;
  }
}

////////////////////////////////////////////////////////////

template<class Indicator>
inline void CutDetector<Indicator>::initHistory(const Indicator& ind) {
  _history = ind;
  _historyExists = true;
}

////////////////////////////////////////////////////////////

template<class Indicator>
bool CutDetector<Indicator>::decide(Similarity sim) {

#ifdef TEST
  cres << "Similarity:\t" << sim << "\tEQ = 0, SIM, UNEQ" << endl;
#endif

  if (sim == SIMILAR) {
    _simCount++;
    if (_simCount > _simParts) {
      _simCount = 0;
      return true;
    }
    else {
      return false;
    }
  }
  else {               // sim != SIMILAR
    _simCount = 0;
  }

  if (sim == UNEQUAL)
    return true;

  return false;        // sim == EQUAL
}

////////////////////////////////////////////////////////////

template<class Indicator>
void CutDetector<Indicator>::adjust(const Indicator& curInd, bool cut) {
  if (!cut)
    _history = _weight * _history + (1 - _weight) * curInd;
  else
    _history = curInd;
}

////////////////////////////////////////////////////////////

template<class Indicator>
ostream& CutDetector<Indicator>::printOn(ostream& os) const {
  os << "[ CutDetector, similar parts = " << _simParts
     << ", current count = " << _simCount
     << ", weight = " << _weight
     << ", history = " << _history << " ]";
  return os;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                 CutDetectorUsingFFT

/**
 * CutDetectorUsingFFT is a CutDetector, which uses frequency
 * spectrum as cut indicator. 
 *
 * @author Christoph Gerum, Silvia Pfeiffer, Peter Tomczyk
 * @version $Id: aucut.h,v 1.13 1998/09/11 11:09:57 tomczyk Exp $
 * @see LinearFFT SpectrumCompare
 */
class CutDetectorUsingFFT : public CutDetector<FFTData> {
  public:
    CutDetectorUsingFFT();
    CutDetectorUsingFFT(unsigned int nparts, Float weight,
      Float eqDev, Float simDev, FFT* fft,
      FFTData::ConvFunc func);
    ~CutDetectorUsingFFT();

    FFT& fftTool() { return *_fftTool; }
    void convFunc(FFTData::ConvFunc convFunc) { _convFunc = convFunc; }
    void deviations(Float eq, Float sim);

    const FFT& fftTool() const { return *_fftTool; }
    FFTData::ConvFunc convFunc() const { return _convFunc; }
    Float eqDeviation() const { return _eqDev; }
    Float simDeviation() const { return _simDev; }

  protected:
    FFTData* analyse(const AudioCache& aucache);
    Similarity compareTo(const FFTData& ind);
    Float euclVecDist(const FFTData& spec1, const FFTData& spec2) const;
    void adjust(const FFTData& spec, bool cut);
    //ostream& printOn(ostream& os) const;

    Float _eqDev;
    Float _simDev;
    FFT* _fftTool;
    FFTData::ConvFunc _convFunc;
    FFTData _tempSpec;
};

////////////////////////////////////////////////////////////

inline
CutDetectorUsingFFT::CutDetectorUsingFFT() 
  : CutDetector<FFTData>(), _eqDev(0.4), _simDev(0.7), 
    _fftTool(new FFT()), _convFunc(&FFTData::amp) {

  if (_fftTool == 0)
    throw Exception("CutDetectorUsingFFT::CutDetectorUsingFFT - cannot allocate memory.");
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
CutDetectorUsingFFT::CutDetectorUsingFFT(unsigned int nparts,
  Float weight, Float eqDev, Float simDev, FFT* fft,
  FFTData::ConvFunc func) : CutDetector<FFTData>(nparts, weight), 
  _eqDev(eqDev), _simDev(simDev), _fftTool(fft), 
  _convFunc(func) {

  if (_fftTool == 0)
    throw Exception("CutDetectorUsingFFT::CutDetectorUsingFFT - cannot allocate memory.");
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
CutDetectorUsingFFT::~CutDetectorUsingFFT() { 
  delete _fftTool; }

#endif

