/*
   This file is part of the Aulib library.
   Name: audata.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "audata.h"
#include "error.h"
#include <algorithm>
#include <sys/stat.h>
using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AudioData

AudioData::AudioData(const string& filename)
  : _filename(filename), _cache(0), _byteReader(0), _byteWriter(0) {

  _file.open(filename.c_str(), ios::binary | ios::in);
  if (!_file)
    throw Exception("AudioData::AudioData() - cannot open file.");

  _dataLocation = 0;
  _dataFormat = DataFormat::mulaw;
  _samplingRate = 8000;
  _bytesInSample = 1;
  _channelCount = 1;
  _info = "No header found.";

  // Calculate data size.
  struct stat stat_buf;
  uint32 filesize = 0;

  if (stat(_filename.c_str(), &stat_buf) != -1)
    filesize = stat_buf.st_size;
  _dataSize = filesize;
}

////////////////////////////////////////////////////////////

AudioData::~AudioData() {
  delete _cache;
  delete _byteReader;
  delete _byteWriter;
}

////////////////////////////////////////////////////////////

size_t AudioData::load(size_t samples) const {

  const uint32 BytesInSample = _bytesInSample;
  Byte bufL[BytesInSample];
  Byte bufR[BytesInSample];
  size_t remain; // number of samples from the current file
                 // position to the end of file
  size_t toLoad; // number of samples will be read

  remain = (_dataLocation + _dataSize - _file.tellg()) /
    (_bytesInSample * _channelCount);
  toLoad = min(samples, remain);

  // before loading samples, assert that there's enough space
  _cache->resize(samples);

  if (_channelCount == 1) { // mono data
    for (size_t i = 0; i < toLoad; i++) {
      if (!_file.read(bufL, BytesInSample))
        throw Exception("AudioFile::load() - error reading file.");
      (*_cache)[i] = _byteReader->toSample(bufL);
    }
  }
  else {                    // stereo data
    for (size_t i = 0; i < toLoad; i++) {
      _file.read(bufL, BytesInSample);
      _file.read(bufR, BytesInSample);
      if (!_file)
        throw Exception("AudioFile::load() - error reading file.");
      (*_cache)[i] = (_byteReader->toSample(bufL) + _byteReader->toSample(bufR)) / 2;
    }
  }

  // fill up the cache
  for (size_t i = toLoad; i < samples; i++)
    (*_cache)[i] = 0;

  return toLoad;
}

////////////////////////////////////////////////////////////

size_t AudioData::loadSecs(size_t secs) const {

  size_t samples, samplesLoaded, secsLoaded;

  samples = secs * _samplingRate * _channelCount;
  samplesLoaded = load(samples);

  // turn samples loaded to seconds
  if (samplesLoaded == samples)
    secsLoaded = secs;
  else
    secsLoaded = samplesLoaded / (_samplingRate * _channelCount);

  return secsLoaded;
}

////////////////////////////////////////////////////////////

void AudioData::store(ofstream& outFile) const {

  const uint32 BytesInSample = _bytesInSample;
  Byte buf[BytesInSample];

  // store samples from cache to outFile
  for (size_t i = 0; i < _cache->size(); i++) {
    _byteWriter->fromSample((*_cache)[i], buf);
    if (!outFile.write(buf, BytesInSample))
      throw Exception("AudioFile::store() - Error writing file.");
    if (_channelCount == 2) {
      if (!outFile.write(buf, BytesInSample))
        throw Exception("AudioFile::store() - Error writing file.");
    }
  }
}

////////////////////////////////////////////////////////////

void AudioData::seek(long int samples, SeekMode mode) const {

  long int offset = samples * _bytesInSample * _channelCount;
  const long int curpos = _file.tellg() - _dataLocation;
  const long int endpos = _dataSize;

  switch (mode) {
    case CUR:
      offset = max(offset, -curpos);
      offset = min(offset, endpos - curpos);
      _file.seekg(offset, ios::cur);
      break;
    case END:
      offset = max(offset, -endpos);
      offset = min(offset, 0L);
      _file.seekg(offset, ios::end);
      break;
    default: // BEG
      offset = max(0L, offset);
      offset = min(offset, endpos);
      _file.seekg(_dataLocation + offset, ios::beg);
  }
}

////////////////////////////////////////////////////////////

ostream& AudioData::printOn(ostream& os) const {

  os << "[ AudioData, file = " << _filename
     << ", sampling rate = " << _samplingRate << " Hz"
     << ", data format = ";

  switch (_dataFormat) {
    case DataFormat::mulaw: os << "8 bit mulaw"; break;
    case DataFormat::lin8:  os << "8 bit linear"; break;
    case DataFormat::lin16: os << "16 bit linear"; break;
    case DataFormat::lin24: os << "24 bit linear"; break;
    case DataFormat::lin32: os << "32 bit linear"; break;
    case DataFormat::IEEEs: os << "IEEE single precision"; break;
    case DataFormat::IEEEd: os << "IEEE double precision"; break;
    case DataFormat::FRAG:  os << "fragmented sampled data"; break;
    default:                os << "unknown";
  }

  os << ", bytes in sample = " << _bytesInSample
     << ", channels = " << _channelCount
     << ", size = " << _dataSize << " bytes"
     << ", info = " << _info << "]";

  return os;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AudioFile

AudioFile::AudioFile(const string& filename)
  : AudioData(filename), _headerID(0) {

  try {
    readHeader();

    // create cache
    _cache = new AudioCache(0, _samplingRate);
    if (_cache == 0)
      throw Exception("AudioFile::AudioFile() - cannot allocate memory.");
  
    // create converter
    switch (_dataFormat) {
      case DataFormat::lin8:
        _byteReader = new Lin8ByteReader;
        _byteWriter = new Lin8ByteWriter;
        break;
      case DataFormat::lin16:
        _byteReader = new Lin16ByteReader;
        _byteWriter = new Lin16ByteWriter;
        break;
      default: // mulaw
        _byteReader = new MulawByteReader;
        _byteWriter = new MulawByteWriter;
    }
    if (_byteReader == 0 || _byteWriter == 0)
      throw Exception("AudioFile::AudioFile() - cannot allocate memory.");
  }

  catch(...) {
    delete _cache;
    delete _byteReader;
    delete _byteWriter;
    throw;
  }
}

////////////////////////////////////////////////////////////

//void AudioFile::playOn(AFPlayer& player, const long length,
//                const int gain, const float delay, 
//                const int device) const {
//  player.play(*this, length, gain, delay, device);
//}

////////////////////////////////////////////////////////////

Byte* AudioFile::swapInt32(Byte cp[4]) const {
  Byte swaptmp;

  swaptmp = cp[0];
  cp[0]   = cp[3];
  cp[3]   = swaptmp;
  swaptmp = cp[1];
  cp[1]   = cp[2];
  cp[2]   = swaptmp;

  return cp;
}

////////////////////////////////////////////////////////////

void AudioFile::readHeader() {

  Byte buf[4];

  // read first 4 Bytes of au_file to decide whether header is there
  if (!_file.read(buf, 4))
    throw Exception("AudioFile::open() - error reading file.");
  
  if (isSNDFile(buf)) { // Header BEGIN
    _headerID = 1;  // file is in .snd format
  
    // read data_location
    if (!_file.read(buf, 4))
      throw Exception("AudioFile::open() - error reading file.");
    _dataLocation = packBytes(buf);

    // read data size
    if (!_file.read(buf, 4))
      throw Exception("AudioFile::open() - error reading file.");
    _dataSize = packBytes(buf);
  
    // read data format
    if (!_file.read(buf, 4))
      throw Exception("AudioFile::open() - error reading file.");
    uint32 format = packBytes(buf);
  
    if (format > DataFormat::UNKNOWN || format < DataFormat::mulaw) {
      throw Exception("AudioFile::open() - unknown data format.");
    }
    else {
      _dataFormat = static_cast<DataFormat::FormatCode>(format);
    }
  
    // read sampling rate
    if (!_file.read(buf, 4))
      throw Exception("AudioFile::open() - error reading file.");
    _samplingRate = packBytes(buf);
  
    // read channel count
    if (!_file.read(buf, 4))
      throw Exception("AudioFile::open() - error reading file.");
    _channelCount = packBytes(buf);
  
    // copy information field
    const int InfoLen = _dataLocation - 23; // 23 bytes have been 
    if (InfoLen > 1) {                      // already read
      char info[InfoLen];
      if (!_file.read(info, InfoLen))
        throw Exception("AudioFile::open() - error reading file.");
      _info.assign(info);
    }
  } // Header END
  
  _bytesInSample = DataFormat::bytesInSample(_dataFormat);
  _file.seekg(_dataLocation, ios::beg);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       WaveFile

WaveFile::WaveFile(const string& filename)
  : AudioData(filename), _fileSize(0), _type(1), _bitsInSample(8) {

  try {
    readHeader();

    // create cache
    _cache = new AudioCache(0, _samplingRate);
    if (_cache == 0)
      throw Exception("AudioFile::AudioFile() - cannot allocate memory.");
  
    // create converter
    switch (_dataFormat) {
      case DataFormat::lin16:
        _byteReader = new Lin16ByteReader;
        _byteWriter = new Lin16ByteWriter;
        break;
      default: // lin8
        _byteReader = new Lin8ByteReader;
        _byteWriter = new Lin8ByteWriter;
    }
    if (_byteReader == 0 || _byteWriter == 0)
      throw Exception("AudioFile::AudioFile() - cannot allocate memory.");
  }

  catch(...) {
    delete _cache;
    delete _byteReader;
    delete _byteWriter;
    throw;
  }
}

////////////////////////////////////////////////////////////

//void WaveFile::playOn(AFPlayer& player, const long length,
//                const int gain, const float delay, 
//                const int device) const {
//  player.play(*this, length, gain, delay, device);
//}

////////////////////////////////////////////////////////////

void WaveFile::readFormatTag(Byte header[2]) {
  if ((header[0] == 0x01) && (header[1] == 0x00)) {
    _type = 1;
    return;
  }
  throw Exception("WaveFile::readFormatTag() - not a WAVE file.");
}

////////////////////////////////////////////////////////////

void WaveFile::readChannelCount(Byte header[2]) {

  if ((header[0] == 0x01) && (header[1] == 0x00)) {
    _channelCount = 1;
    return;
  }

  if ((header[0] == 0x02) && (header[1] == 0x00)) {
    _channelCount = 2;
    return;
  }

  throw Exception("WaveFile::readChannelCount() - cannot handle more than 2 channels.");
}

////////////////////////////////////////////////////////////

void WaveFile::readSamplingRate(Byte header[2]) {

  if ((header[0] == 0x11) && (header[1] == 0x2B)
      && (header[2] == 0x00) && (header[3] == 0x00)) {
    _samplingRate = 11025;
    return;
  }

  if ((header[0] == 0x22) && (header[1] == 0x56)
      && (header[2] == 0x00) && (header[3] == 0x00)) {
    _samplingRate = 22050;
    return;
  }

  if ((header[0] == 0x44) && (header[1] == 0xAC)
      && (header[2] == 0x00) && (header[3] == 0x00)) {
    _samplingRate = 44100;
    return;
  }

  throw Exception("WaveFile::readSamplingRate() - unknown sampling rate.");
}

////////////////////////////////////////////////////////////

void WaveFile::readBytesPerSample(Byte header[2]) {

  if ((header[0] == 0x01) && (header[1] == 0x00)) {
    _bytesInSample = 1;
    _dataFormat = DataFormat::lin8;
    return;
  }

  if ((header[0] == 0x02) && (header[1] == 0x00) && (_channelCount == 1)) {
    _dataFormat = DataFormat::lin16;
    _bytesInSample = 2;
    return;
  }

  if ((header[0] == 0x02) && (header[1] == 0x00) && (_channelCount == 2)) {
    _dataFormat = DataFormat::lin8;
    _bytesInSample = 1;
    return;
  }

  if ((header[0] == 0x04) && (header[1] == 0x00)) {
    _dataFormat = DataFormat::lin16;
    _bytesInSample = 2;
    return;
  }

  throw Exception("WaveFile::readBytesPerSample() - unknown data format.");
}

////////////////////////////////////////////////////////////

void WaveFile::readBitsPerSample(Byte header[2]) {

  if ((header[0] == 0x08) && (header[1] == 0x00)) {
    _bitsInSample = 8;
    return;
  }

  if ((header[0] == 0x12) && (header[1] == 0x00)) {
    _bitsInSample = 8;
    return;
  }

  if ((header[0] == 0x16) && (header[1] == 0x00)) {
    _bitsInSample = 8;
    return;
  }

  throw Exception("WaveFile::readBitInSample() - unknown data format.");
}

////////////////////////////////////////////////////////////

void WaveFile::readHeader() {

  Byte buf[4];

  // read 4 bytes - it must be the RIFF header
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  if (!isRiffHeader(buf))
    throw Exception("WaveFile::open() - RIFF header not found.");

  // read 4 bytes - it's the file size less 8 bytes
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  _fileSize = packBytes(buf) + 8;

  // read 4 bytes - it must be the WAVE header
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  if (!isWaveHeader(buf))
    throw Exception("WaveFile::open() - WAVE header not found.");

  // read 4 bytes - it must be the fmt header
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  if (!isFmtHeader(buf))
    throw Exception("WaveFile::open() - fmt header not found.");

  // read 4 bytes - it's the length of the format chunk
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  if (!isOrdinaryFormatChunk(buf))
    throw Exception("WaveFile::open() - cannot handle such data chunks.");

  // read 2 bytes - it's the format tag
  if (!_file.read(buf, 2))
    throw Exception("WaveFile::open() - error reading file.");
  readFormatTag(buf);

  // read 2 bytes - it's the number of channels
  if (!_file.read(buf, 2))
    throw Exception("WaveFile::open() - error reading file.");
  readChannelCount(buf);

  // read 4 bytes - it's the sampling rate
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  readSamplingRate(buf);

  // read 4 bytes - it's the number of bytes per second
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  readBytesPerSecond(buf);

  // read 2 bytes - it's the number of bytes per sample
  if (!_file.read(buf, 2))
    throw Exception("WaveFile::open() - error reading file.");
  readBytesPerSample(buf);

  // read 2 bytes - it's the number of bits per sample
  if (!_file.read(buf, 2))
    throw Exception("WaveFile::open() - error reading file.");
  readBitsPerSample(buf);

  // read 4 bytes - it must be the data header
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  if (!isDataHeader(buf))
    throw Exception("WaveFile::open() - unknown header.");

  // read 4 bytes - it's the size of the data chunk
  if (!_file.read(buf, 4))
    throw Exception("WaveFile::open() - error reading file.");
  _dataSize = packBytes(buf);

  // here begins the audio data - set the pointer
  _dataLocation = _file.tellg();

  // write info field
  _info.assign("WAVE file.");
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                    AudioFactory

AudioData* AudioFactory::openFile(string filename) {

  if (isAuFile(filename))
    return (new AudioFile(filename));

  if (isWavFile(filename))
    return (new WaveFile(filename));

  // otherwise
  throw Exception("AudioFactory::openFile() - file extention unknown.");
}

////////////////////////////////////////////////////////////

bool AudioFactory::isAuFile(string filename) {

  string suffix;
  string::size_type pos;

  pos = filename.find_last_of('.');

  if (pos != string::npos) {
    suffix = filename.substr(pos);
    if (suffix == ".au")
      return true;
  }
  return false;
}

////////////////////////////////////////////////////////////

bool AudioFactory::isWavFile(string filename) {

  string suffix;
  string::size_type pos;

  pos = filename.find_last_of('.');
  if (pos != string::npos) {
    suffix = filename.substr(pos);
    if (suffix == ".wav")
      return true;
  }
  return false;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       ByteReader

const Sample MulawByteReader::ulaw2lin[256] = {
 -32124, -31100, -30076, -29052, -28028, -27004, -25980, -24956,
 -23932, -22908, -21884, -20860, -19836, -18812, -17788, -16764,
 -15996, -15484, -14972, -14460, -13948, -13436, -12924, -12412,
 -11900, -11388, -10876, -10364,  -9852,  -9340,  -8828,  -8316,
  -7932,  -7676,  -7420,  -7164,  -6908,  -6652,  -6396,  -6140,
  -5884,  -5628,  -5372,  -5116,  -4860,  -4604,  -4348,  -4092,
  -3900,  -3772,  -3644,  -3516,  -3388,  -3260,  -3132,  -3004,
  -2876,  -2748,  -2620,  -2492,  -2364,  -2236,  -2108,  -1980,
  -1884,  -1820,  -1756,  -1692,  -1628,  -1564,  -1500,  -1436,
  -1372,  -1308,  -1244,  -1180,  -1116,  -1052,   -988,   -924,
   -876,   -844,   -812,   -780,   -748,   -716,   -684,   -652,
   -620,   -588,   -556,   -524,   -492,   -460,   -428,   -396,
   -372,   -356,   -340,   -324,   -308,   -292,   -276,   -260,
   -244,   -228,   -212,   -196,   -180,   -164,   -148,   -132,
   -120,   -112,   -104,    -96,    -88,    -80,    -72,    -64,
    -56,    -48,    -40,    -32,    -24,    -16,     -8,      0,
  32124,  31100,  30076,  29052,  28028,  27004,  25980,  24956,
  23932,  22908,  21884,  20860,  19836,  18812,  17788,  16764,
  15996,  15484,  14972,  14460,  13948,  13436,  12924,  12412,
  11900,  11388,  10876,  10364,   9852,   9340,   8828,   8316,
   7932,   7676,   7420,   7164,   6908,   6652,   6396,   6140,
   5884,   5628,   5372,   5116,   4860,   4604,   4348,   4092,
   3900,   3772,   3644,   3516,   3388,   3260,   3132,   3004,
   2876,   2748,   2620,   2492,   2364,   2236,   2108,   1980,
   1884,   1820,   1756,   1692,   1628,   1564,   1500,   1436,
   1372,   1308,   1244,   1180,   1116,   1052,    988,    924,
    876,    844,    812,    780,    748,    716,    684,    652,
    620,    588,    556,    524,    492,    460,    428,    396,
    372,    356,    340,    324,    308,    292,    276,    260,
    244,    228,    212,    196,    180,    164,    148,    132,
    120,    112,    104,     96,     88,     80,     72,     64,
     56,     48,     40,     32,     24,     16,      8,      0 };

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       ByteWriter

Byte MulawByteWriter::lin2ulaw(Sample sample) const {
  int sign, exponent, mantissa;
  Byte ulawbyte;

  // Get the sample into sign-magnitude.
  sign = (sample >> 8) & 0x80;            // set aside the sign
  if ( sign != 0 ) sample = -sample;      // get magnitude
  if ( sample > 32635 ) sample = 32635;   // clip the magnitude

  // Convert from 16 bit linear to ulaw.
  sample = sample + 0x84;
  exponent = expLut[( sample >> 7 ) & 0xFF];
  mantissa = ( sample >> ( exponent + 3 ) ) & 0x0F;
  ulawbyte = ~ ( sign | ( exponent << 4 ) | mantissa );

  return ulawbyte;
}

const int MulawByteWriter::expLut[256] = {
 0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
 4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
 5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
 5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};

