/*
   This file is part of the Aulib library.
   Name: bg.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "bg.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                 BGFinder

BGFinder::BGFinder(size_t size, size_t step, Float limit)
  : _winSize(size), _winStep(step), _relLimit(limit) {

  if (_winStep == 0)
    throw Exception("BackgroundExtractor::BackgroundExtractor() - step cannot be 0.");
  if (_relLimit < 0 || _relLimit > 1)
    throw Exception("BackgroundExtractor::BackgroundExtractor() - relLimit must be of [0,1].");
}

//////////////////////////////////////////////////////////////

vector<BGFinder::Loudness> BGFinder::calcValues(const AudioData& audata) const {

  vector<Loudness> values;
  size_t pos = 0;

  audata.seek(pos, AudioData::BEG);

  while(audata.load(_winSize) == _winSize) {
    values.push_back(opLoudness(audata.cache()));
    pos += _winStep;
    audata.seek(pos, AudioData::BEG);
  }
  return values;
}

//////////////////////////////////////////////////////////////

vector<Segment> BGFinder::calcPositions(const vector<Loudness>& values) const {

  const Loudness LIMIT = absLimit(values);
  size_t begin = 0; // segment bounds
  size_t end   = 0;
  size_t pos   = 0;
  vector<Segment> segments;


  for (size_t i = 0; i < values.size(); i++) {
    if (values[i] < LIMIT) {  // quiet window found
      if (pos > end) {        // begin of a new segment
        if (end != 0)         // store previous segment if not (0, 0)
          segments.push_back(Segment(begin, end));
        begin = pos;          // set segment bounds
        end = pos + _winSize;
      }
      else {                  // it is still the old segment
        end = pos + _winSize; // move segment end
      }
    }
    pos += _winStep;
  }

  // if the last segment has not been stored yet, do it now
  if (end != 0 && (segments.size() == 0 || segments.back().end != end))
    segments.push_back(Segment(begin, end));

  return segments;
}

//////////////////////////////////////////////////////////////

ostream& BGFinder::printOn(ostream& os) const {
  os << "[ BGFinder, ... ]";
  return os;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                 BGAnalyser


Float BGAnalyser::calcNorm(const vector<Segment>& segments) const {
  size_t length = 0;
  for (size_t i = 0; i < segments.size(); i++)
    length += segments[i].end - segments[i].begin;
  return (1.0 * _winStep / (length + _winStep - _fftTool->winSize()));
}

//////////////////////////////////////////////////////////////

FFTData BGAnalyser::analyse(const AudioData& audata,
  const vector<Segment>& segments) const {

  FFTData cumSpec(_fftTool->winSize()); // result
  Float norm = calcNorm(segments);      // normalize factor
  pos_t pos;                            // file position

  // if FFT window size changed, change size of _tempSpec
  if (_tempSpec.size() != _fftTool->winSize())
    _tempSpec.resize(_fftTool->winSize());

  for (size_t i = 0; i < segments.size(); i++) {
    pos = segments[i].begin;
    while ((pos + _fftTool->winSize()) < segments[i].end) {
      audata.seek(pos);
      audata.load(_fftTool->winSize());
      updateSpec(audata.cache(), &cumSpec, norm);
      pos += _winStep;
    }
  }
  return cumSpec;
}

//////////////////////////////////////////////////////////////

ostream& BGAnalyser::printOn(ostream& os) const {
  os << "[ BGAnalyser, ... ]";
  return os;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                 BGExtractor

void BGExtractor::extract(const AudioCache& in, AudioCache* out) const {
  _fftTool->fft(in, &_tempSpec);
  for (size_t i = 0; i < _fftTool->winSize(); i++) {
    if (!_freqs[i])
      _tempSpec[i] = 0.0;
  }
  _fftTool->invfft(&_tempSpec);

  // now _tempSpec contains the filtered wave
  copyIntoCache(_tempSpec, out);
}

//////////////////////////////////////////////////////////////


ostream& BGExtractor::printOn(ostream& os) const {
  os << "[ BackgroundExtractor, ... ]";
  return os;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      BGUtil

vector<bool> BGUtil::bgFreqs(const FFTData& spec, FFTData::ConvFunc func, Float relLimit) {

  vector<bool> freqs(spec.size());

  // Calculate minimum amplitude a frequency must have to be
  // a background frequency.
  FFTData::CIter itr = max_element(spec.cbegin(func), spec.cend(func));
  Float minAmp = (itr == spec.cend(func) ? 0 : *itr * relLimit);

  // Mark frequencies.
  for (size_t i = 0; i < spec.size(); i++)
    freqs[i] = ((spec.*func)(i) > minAmp);

  return freqs;
}

