/*
   This file is part of the Aulib library.
   Name: fft.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "fft.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       FFTUtil

size_t FFTUtil::samplesInBits(size_t samples) {
  size_t bits = 0;
  for (size_t i = samples >> 1; i != 0; i >>= 1)
    bits++;
  return bits;
}

////////////////////////////////////////////////////////////

size_t freqIndex(size_t freq, size_t nfreqs, uint32 sampRate) {
  const double fdist = 1.0 * sampRate / nfreqs;
  for (size_t i = 0; i < nfreqs; i++) {
    if (i * fdist >= freq)
      return i;
  }
  return nfreqs;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       FFTData

FFTData operator+(const FFTData& data1, const FFTData& data2) {
  if (data1.size() != data2.size())
    throw Exception("operator+ - arguments have different size.");
  FFTData result(data1.size());
  for (FFTData::size_type i = 0; i < result.size(); i++)
    result[i] = data1[i] + data2[i];
  return result;
}

FFTData operator-(const FFTData& data1, const FFTData& data2) {
  if (data1.size() != data2.size())
    throw Exception("operator+ - arguments have different size.");
  FFTData result(data1.size());
  for (FFTData::size_type i = 0; i < result.size(); i++)
    result[i] = data1[i] - data2[i];
  return result;
}

FFTData operator*(Float x, const FFTData& data) {
  FFTData result(data.size());
  for (FFTData::size_type i = 0; i < result.size(); i++)
    result[i] = x * data[i];
  return result;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       BaseFFT

BaseFFT::BaseFFT(size_type winSize, WinFunc winFunc)
  : _bits(FFTUtil::samplesInBits(winSize)), 
    _winSize(FFTUtil::bitsInSamples(_bits)),
    _winFunc(winFunc), _sineTable(0), _cosineTable(0), _winTable(0) {

  if (_winSize != 0) {
    createTables();
    setTables();
  }
}

/////////////////////////////////////////////////////////////////

void BaseFFT::set(size_type winSize, WinFunc winFunc) {

  _winFunc = winFunc;
  if (winSize != _winSize) {
    _bits = FFTUtil::samplesInBits(winSize);
    _winSize = FFTUtil::bitsInSamples(_bits);
    deleteTables();
    createTables();
  }
  setTables();
}

/////////////////////////////////////////////////////////////////

void BaseFFT::createTables() {

  _sineTable   = new Float[_winSize];
  _cosineTable = new Float[_winSize];
  _winTable    = new Float[_winSize];
  _sigma       = new size_type[_winSize];

  if (_sineTable == 0 || _cosineTable == 0 
      || _winTable == 0 || _sigma == 0) {
    deleteTables();
    throw Exception("BaseFFT::BaseFFT() - cannot allocate memory.");
  }
}

/////////////////////////////////////////////////////////////////

void BaseFFT::setTables() {

  if (_winSize == 0)
    return;

  const Float TWOPIoN   = (atan(1.0) * 8.0) / _winSize;
  const Float TWOPIoNm1 = (atan(1.0) * 8.0) / (_winSize - 1);

  for (size_type i = 0; i < _winSize; i++) {
    _sigma[i]       = reverse(i);
    _sineTable[i]   = sin(i * TWOPIoN);
    _cosineTable[i] = cos(i * TWOPIoN);

    // window functions:
    switch (_winFunc) {
      case NONE:
        _winTable[i] = 1.0;
        break;
      case HANNING:
        _winTable[i] = 0.50 + (0.50 * cos (TWOPIoNm1 * (i - _winSize/2)));
        break;
      default: // Hamming
        _winTable[i] = 0.54 + (0.46 * cos (TWOPIoNm1 * (i - _winSize/2)));
    }
  }
}

/////////////////////////////////////////////////////////////////

void BaseFFT::fft(Container* wave) const {

  if (wave == 0 || wave->_size < _winSize)
    throw Exception("BaseFFT::fft() - bad argument.");

  scale(wave);
  window(wave);
  fftAlgorithm(wave);
  reorder(wave);
}

/////////////////////////////////////////////////////////////////

void BaseFFT::invfft(Container* wave) const {

  if (wave == 0 || wave->_size < _winSize)
    throw Exception("BaseFFT::invfft() - bad argument.");

  invfftAlgorithm(wave);
  reorder(wave);
  unwindow(wave);
}

/////////////////////////////////////////////////////////////////

void BaseFFT::fft(const AudioCache& wave, Container* spec) const {

  if (spec == 0 || spec->_size < _winSize)
    throw Exception("BaseFFT::fft() - bad argument.");

  // Copy 'wave' into 'spec'. Perform FFT on 'spec'.
  copy(wave, spec);
  scale(spec);
  window(spec);
  fftAlgorithm(spec);
  reorder(spec);
}

/////////////////////////////////////////////////////////////////

void BaseFFT::copy(const AudioCache& cache, Container* spec) const {

  size_type i = 0;

  for (; i < spec->_size && i < cache.size(); i++)
    spec->_data[i] = 1.0 * cache[i];
  for (; i < spec->_size; i++)
    spec->_data[i] = 0.0;
}

/////////////////////////////////////////////////////////////////

void BaseFFT::fftAlgorithm(Container* wave) const {

  int       loop, loop1, loop2;
  unsigned	i1; // going to right shift this
  int		i2, i3, i4, y;
  Float		a1, a2, b1, b2, z1, z2;

  i1 = _winSize / 2;
  i2 = 1;

  // perform the butterfly's
  for (loop = 0; loop < static_cast<int>(_bits); loop++) {
    i3 = 0;
    i4 = i1;

    for (loop1 = 0; loop1 < i2; loop1++) {
      y  = permute(i3 / static_cast<int>(i1));
      z1 = _cosineTable[y];
      z2 = -_sineTable[y];

      for (loop2 = i3; loop2 < i4; loop2++) {
        a1 = wave->_data[loop2].real();
        a2 = wave->_data[loop2].imag();

        b1 = z1 * wave->_data[loop2+i1].real() - z2 * wave->_data[loop2+i1].imag();
        b2 = z2 * wave->_data[loop2+i1].real() + z1 * wave->_data[loop2+i1].imag();

        wave->_data[loop2] = Container::value_type(a1 + b1, a2 + b2);
        wave->_data[loop2+i1] = Container::value_type(a1 - b1, a2 - b2);
      }
      i3 += (i1 << 1);
      i4 += (i1 << 1);
    }
    i1 >>= 1;
    i2 <<= 1;
  }
}


/////////////////////////////////////////////////////////////////

void BaseFFT::invfftAlgorithm(Container* wave) const {

  int       loop, loop1, loop2;
  unsigned	i1; // going to right shift this
  int		i2, i3, i4, y;
  Float		a1, a2, b1, b2, z1, z2;

  i1 = _winSize / 2;
  i2 = 1;

  // perform the butterfly's
  for (loop = 0; loop < static_cast<int>(_bits); loop++) {
    i3 = 0;
    i4 = i1;

    for (loop1 = 0; loop1 < i2; loop1++) {
      y  = permute(i3 / static_cast<int>(i1));
      z1 = _cosineTable[y];
      z2 = _sineTable[y];

      for (loop2 = i3; loop2 < i4; loop2++) {
        a1 = wave->_data[loop2].real();
        a2 = wave->_data[loop2].imag();

        b1 = z1 * wave->_data[loop2+i1].real() - z2 * wave->_data[loop2+i1].imag();
        b2 = z2 * wave->_data[loop2+i1].real() + z1 * wave->_data[loop2+i1].imag();

        wave->_data[loop2] = Container::value_type(a1 + b1, a2 + b2);
        wave->_data[loop2+i1] = Container::value_type(a1 - b1, a2 - b2);
      }
      i3 += (i1 << 1);
      i4 += (i1 << 1);
    }
    i1 >>= 1;
    i2 <<= 1;
  }
}

/////////////////////////////////////////////////////////////////

const BaseFFT::size_type BaseFFT::reverse(size_type i) const {

  size_type res  = 0;
  size_type bits = _bits;

  while (bits--) {
    res <<= 1;
    res |= (i & 1);
    i >>= 1;
  }
  return res;
}

/////////////////////////////////////////////////////////////////

void BaseFFT::reorder(Container* wave) const {

  Container::value_type tmp;
  size_type j;

  for (size_type i = 0; i < _winSize; i++) {
    j = permute(i);
    if (i > j) {   // So we only do each pair once
      tmp = wave->_data[i];
      wave->_data[i] = wave->_data[j];
      wave->_data[j] = tmp;
    }
  }
}

/////////////////////////////////////////////////////////////////

ostream& BaseFFT::printOn(ostream& os) const {
  os << "[ FFT, window size = " << _winSize
    << "(" << _bits << " bits), windowing function = ";
  switch (_winFunc) {
    case HAMMING: os << "Hamming"; break;
    case HANNING: os << "Hanning"; break;
    default: os << "none";
  }
  os << "]";
  return os;
}

