/*
   This file is part of the Aulib library.
   Name: rangeTableVectorFile.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include <iostream.h>
#include <fstream.h>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include "byteOrder.h"
#include "vectorFile.H"
#include "rangeTableVectorFile.H"

#ifndef _WIN32
#ifdef DAMN_INTEL_BYTE_ORDER
//#include <sys/mode.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#endif not _WIN32

// ---------------------------------------------------------------------
// -  CLASS rangeTableVectorFile
// ---------------------------------------------------------------------
int
rangeTableVectorFile::newDB(string _filename,
			    int _startFrameNo, int _stopFrameNo,
			    int _size)
{
  filename = _filename;
  datafilename = _filename+".vec";
  startFrameNo = _startFrameNo;
  stopFrameNo = _stopFrameNo;
  size = _size;
  minValue=FLT_MAX;
  maxValue=-FLT_MAX;

  // open vector data file
  if ( (file = fopen(datafilename.c_str(),"wb")) == NULL) {
    cerr << "Cannot create DB file "
         << datafilename.c_str()
         << "." << endl;
    return false;
  }

  // open range table index file
  ofstream fileRT(filename.c_str());

  if ( !fileRT) {
    cerr << "Cannot create DB fileRT "
	 << filename.c_str()
	 << "." << endl;
    return false;
  }

  // write header to index file
  fileRT << startFrameNo
	 << "\t"
	 << stopFrameNo
	 << endl;
  fileRT << size
	 << "\t"
	 << minValue
	 << "\t"
	 << maxValue
	 << endl;

  fileRT.close();

#ifndef _WIN32
  chmod (filename.c_str(),S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
  chmod (datafilename.c_str(),S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
#endif not _WIN32

  // reopen file
  fclose(file);
  file = fopen(datafilename.c_str(),"r+b");

  return true;
}


int
rangeTableVectorFile::openDB(string _filename,
			     int _startFrameNo, int _stopFrameNo,
			     int _size)
{
  filename = _filename;
  datafilename = _filename+".vec";
  startFrameNo = _startFrameNo;
  stopFrameNo = _stopFrameNo;
  size = _size;

  // open range table index file
  ifstream fileRT(filename.c_str());
  if (!fileRT) {
    return newDB(filename, startFrameNo, stopFrameNo);
  }

  // check, whether DB file has write size
  int dbstartFrameNo, dbstopFrameNo, dbsize;
  fileRT >> dbstartFrameNo
	 >> dbstopFrameNo
	 >> dbsize
	 >> minValue
	 >> maxValue;
  fileRT.close();

  if ( (dbstartFrameNo != startFrameNo) ||
       (dbstopFrameNo != stopFrameNo) ||
       (dbsize != size) ) {
    // DB has wrong size, generate new
    if ((size == -1) && (dbstartFrameNo == starFrameNo) &&
	(dbstopFrameNo == dbstopFrameNo)) {
      size = dbsize;
    } else {
      cerr << "Tried to open DB "
           << filename.c_str()
           << " with wrong parameters!" << endl;
      return false;
      //    return newDB(filename, startFrameNo, stopFrameNo, size);
    }
  }

  // open vector data file
  if (file) {
    fclose(file);
    file = NULL;
  }
  file = fopen(datafilename.c_str(),"r+b");
  if (!file) {
    cerr << "Warning: "
	 << datafilename.c_str()
	 << " does not exist: creating new!" << endl;
    if ( (file = fopen(datafilename.c_str(),"wb")) == NULL) {
      cerr << "Cannot create DB file "
	   << datafilename.c_str()
	   << "." << endl;
      return false;
    }
    fclose(file);
    file = fopen(datafilename.c_str(),"r+b");
  }

  return true;
}

int
rangeTableVectorFile::closeDB()
{
  if (file) fclose(file);
  file = NULL;
  return 0;
}

int
rangeTableVectorFile::save(rangeTableVectorType value)
{
  // save range table index file
  ofstream fileRT(filename.c_str());
  if ( !fileRT) {
    cerr << "Cannot create DB fileRT "
	 << filename.c_str()
	 << "." << endl;
    return false;
  }

  // write header
  fileRT << startFrameNo
	 << "\t"
	 << stopFrameNo
	 << endl;
  fileRT << size
	 << "\t"
	 << minValue
	 << "\t"
	 << maxValue
	 << endl;
  fileRT << value;
  fileRT.close();

  // vector data is saved to file constantly during use

  return true;
}


rangeTableVectorType
rangeTableVectorFile::load(int _startFrameNo, int _stopFrameNo)
{
  rangeTableVectorType value;

  ifstream fileRT(filename.c_str());
  if (!fileRT) return value;

  // load value
  int dbstartFrameNo, dbstopFrameNo, dbsize, dbmaxValue, dbminValue;
  fileRT >> dbstartFrameNo
	 >> dbstopFrameNo
	 >> dbsize
	 >> dbmaxValue
	 >> dbminValue
	 >> value;
  fileRT.close();

  return value;
}

rangeTableVectorFile::rangeTableVectorFile () {file=NULL; startFrameNo=-1; stopFrameNo=-1;}
rangeTableVectorFile::~rangeTableVectorFile () { if (file) fclose(file); }
